module Ocache
  class Proxy < BasicObject
    private
    def initialize(module_target, key_name=nil, module_options={})
      @module_target = module_target

      if cache_options?(key_name)
        @module_options = key_name
      else
        @key_name = key_name
        @module_options = module_options
      end
    end

    def method_missing(method_target, *method_arguments)
      @method_target = method_target
      @method_arguments = method_arguments

      @key_name = generate_cache_key unless @key_name

      ::Rails.cache.fetch(@key_name, @module_options) do
        @module_target.__send__(method_target, *method_arguments)
      end
    end

    def generate_cache_key
      if cache_instance_attributes?(@module_target)
        key_module_target = @module_target.cache_key
      elsif @module_target.respond_to?(:model_name)
        key_module_target = @module_target.model_name.cache_key
      else
        key_module_target = @module_target
      end

      "#{key_module_target}/#{@method_target}/#{expand_cache_key(@method_arguments)}".chomp('/')
    end

    def cache_options?(possible_options)
      %w(force expires_in race_condition_ttl raw compress compress_threshold namespace).any? { |option|
        possible_options.inspect.include? option
      }
    end

    def cache_instance_attributes?(possible_object)
      possible_object.respond_to?(:id) || possible_object.respond_to?(:created_at) || possible_object.respond_to?(:updated_at)
    end

    def expand_cache_key(arguments)
      ::ActiveSupport::Cache.expand_cache_key(arguments)
    end
  end
end
